/* adc_driver.c - adc single master driver

Copyright (c) 2012 David Michael Betz

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include <propeller.h>
#include "adc_driver.h"
#include "i2c_driver.h"
#include "i2c_driver_lib.h"

#ifndef TRUE
#define TRUE    1
#define FALSE   0
#endif

/* adc state information */
static _COGMEM volatile ADC_MAILBOX *mailbox;
static int addr, mask;

static int adcStartRead(int addr, int mask);

_NAKED int main(void)
{
    ADC_INIT *init;
    ADC_CMD cmd;
    
    /* get the mailbox address and the initialization structure */
    mailbox = (ADC_MAILBOX *)PAR;
    init = mailbox->init;
    
    /* get the COG initialization parameters */
    i2cInit(init->scl, init->sda, init->ticks_per_cycle);
    addr = (int)init->addr;
    mask = (int)init->mask;
    
    /* empty the circular buffer */
    mailbox->iPtr = mailbox->oPtr = 0;
    
    /* tell the caller that we're done with initialization */
    mailbox->cmd = ADC_CMD_IDLE;
    
    /* get ready to sample the ADC */
    if (adcStartRead(addr, mask) != I2C_OK)
        return 1;
    
    /* handle ADC input and mailbox requests */
    for (;;) {
        int byte1, byte2;
        uint32_t sts;
    
        /* check for the next request */
        if ((cmd = mailbox->cmd) != ADC_CMD_IDLE) {
            switch (cmd) {
            case ADC_CMD_FLUSH:
                i2cStop();
                adcStartRead(addr, mask);
                mailbox->iPtr = mailbox->oPtr = 0;
                break;
            default:
                sts = ADC_ERR_UNKNOWN_CMD;
                break;
            }
        
            /* finish request and return the result */
            mailbox->sts = sts;
            mailbox->cmd = ADC_CMD_IDLE;
        }
        
        /* receive a sample from the ADC */
        byte1 = i2cReceiveByte(TRUE);
        byte2 = i2cReceiveByte(TRUE);
        if (byte1 < 0 || byte2 < 0)
            return 1;
        
        /* lock the circular buffer */
        while (lockset(mailbox->lock) != 0)
            ;
        
        /* store the sample */
        mailbox->buffer[mailbox->iPtr++] = (byte1 << 8) | byte2;
        if (mailbox->iPtr >= mailbox->size)
            mailbox->iPtr = 0;
        
        /* check for buffer overflow and keep the newest samples */
        if (mailbox->iPtr == mailbox->oPtr) {
            if (++mailbox->oPtr >= mailbox->size)
                mailbox->oPtr = 0;
        }
        
        /* unlock the circular buffer */
        lockclr(mailbox->lock);
    }
    
    return 0;
}

static int adcStartRead(int addr, int mask)
{
    /* set the register address and the command bits */
    i2cStart();
    if (i2cSendByte(addr | I2C_WRITE) != I2C_OK)
        return I2C_ERR_SEND_HDR;
    if (i2cSendByte(mask << 4) != I2C_OK)
        return I2C_ERR_SEND;
    
    /* start reading the conversion results register */
    i2cStart();
    return i2cSendByte(addr | I2C_READ);
}

